#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

import numpy as np
from numpy.random import multivariate_normal, permutation
import pandas as pd
from pandas import DataFrame
import matplotlib.pyplot as plt

import tensorflow as tf
from tensorflow.keras import layers, models

import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


print('\nSet a random seed')
np.random.seed(20190220)
tf.random.set_seed(20190220)

print('\nGenerate a training dateset.')
n0, mu0, variance0 = 20, [10, 11], 20
data0 = multivariate_normal(mu0, np.eye(2)*variance0, n0)
df0 = DataFrame(data0, columns=['x1', 'x2'])
df0['t'] = 0

n1, mu1, variance1 = 15, [18, 20], 22
data1 = multivariate_normal(mu1, np.eye(2)*variance1, n1)
df1 = DataFrame(data1, columns=['x1', 'x2'])
df1['t'] = 1

df = pd.concat([df0, df1], ignore_index=True)
train_set = df.reindex(permutation(df.index)).reset_index(drop=True)

print('\nThe training dataset consists of coordinates (x1,x2) and label values t=0,1.')
print(train_set)

plt.scatter(df1['x1'], df1['x2'], marker='x')
plt.scatter(df0['x1'], df0['x2'], marker='o')
plt.savefig('ex2-1a.png')
if flag_fig: plt.show()

print('\nStore the coordinates $(x_1,x_2)$ and label values $t$ into NumPy arrays.')
train_x = train_set[['x1', 'x2']].values
train_t = train_set['t'].values

print('\nDefine a model to predict the probability of `t=1`.')
model = models.Sequential()
model.add(layers.Dense(1, activation='sigmoid', input_shape=(2,),
                       name='logistic_regression'))
model.summary()

print('\nCompile the model using the Adam optimizer, and Cross entroy as a loss function.')
model.compile(optimizer='adam',
              loss='binary_crossentropy',
              metrics=['acc'])

print('\nApply the optimization for 10 epochs.')
history = model.fit(train_x, train_t,
                    batch_size=len(train_x), epochs=10)

print('\nApply the optimization for additional 5,000 epochs.')
time_start = time.perf_counter()
history = model.fit(train_x, train_t,
                    batch_size=len(train_x), epochs=5000, verbose=0)
time_end = time.perf_counter()
print('time elapsed =', time_end-time_start)

print('\nPlot charts for accuracy and loss.')
DataFrame({'acc': history.history['acc']}).plot()
plt.savefig('ex2-1b.png')
if flag_fig: plt.show()
DataFrame({'loss': history.history['loss']}).plot()
plt.savefig('ex2-1c.png')
if flag_fig: plt.show()

print('\nShow weights after the training.')
print(model.get_weights())

print('\nPlot a chart using the final weights to show the result.')
[[w1], [w2]], [b] = model.get_weights()

train_set0 = train_set[train_set['t'] == 0]
train_set1 = train_set[train_set['t'] == 1]

fig = plt.figure(figsize=(7, 7))
subplot = fig.add_subplot(1, 1, 1)
subplot.set_ylim([0, 30])
subplot.set_xlim([0, 30])
subplot.scatter(train_set1.x1, train_set1.x2, marker='x')
subplot.scatter(train_set0.x1, train_set0.x2, marker='o')

xs = np.linspace(0, 30, 10)
ys = - (w1*xs/w2 + b/w2)
subplot.plot(xs, ys)

field = [[(1 / (1 + np.exp(-(w1*x1 + w2*x2 + b))))
          for x1 in np.linspace(0, 30, 100)]
         for x2 in np.linspace(0, 30, 100)]
subplot.imshow(field, origin='lower', extent=(0, 30, 0, 30),
               vmin=0, vmax=1, cmap=plt.cm.gray_r, alpha=0.5)
plt.savefig('ex2-1.png')
if flag_fig: plt.show()
