#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

# Double Layer Network Example

import numpy as np
from numpy.random import multivariate_normal, permutation
import pandas as pd
from pandas import DataFrame
import matplotlib.pyplot as plt

import tensorflow as tf
from tensorflow.keras import layers, models, initializers

import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


np.random.seed(20190830)
tf.random.set_seed(20190703)
# np.random.seed(5)
# tf.random.set_seed(5)

print('####[DNE-02]\nGenerate a training dateset.')
def generate_datablock(n, mu, var, t):
    data = multivariate_normal(mu, np.eye(2)*var, n)
    df = DataFrame(data, columns=['x1', 'x2'])
    df['t'] = t
    return df

df0 = generate_datablock(30, [-7, -7], 18, 1)
df1 = generate_datablock(30, [-7, 7], 18, 0)
df2 = generate_datablock(30, [7, -7], 18, 0)
df3 = generate_datablock(30, [7, 7], 18, 1)

plt.scatter(df3['x1'], df3['x2'], marker='x', color='blue')
plt.scatter(df2['x1'], df2['x2'], marker='o', color='salmon')
plt.scatter(df1['x1'], df1['x2'], marker='o', color='tomato')
plt.scatter(df0['x1'], df0['x2'], marker='x', color='slateblue')
plt.savefig('ex3-4a.png')
if flag_fig: plt.show()

df = pd.concat([df0, df1, df2, df3], ignore_index=True)
train_set = df.reindex(permutation(df.index)).reset_index(drop=True)

print('####[DNE-03]\nStore the coordinates $(x_1,x_2)$ and label values $t$ into NumPy arrays.')
train_x = train_set[['x1', 'x2']].values
train_t = train_set['t'].values


print('####[DNE-04]\nDefine a model for the binary classification using two hidden layers.\n\nWeights in the hidden layers are initialized with random values.')
model = models.Sequential()
model.add(layers.Dense(2, activation='tanh', input_shape=(2,),
                       kernel_initializer=initializers.TruncatedNormal(),
                       name='hidden1'))
model.add(layers.Dense(2, activation='tanh',
                       kernel_initializer=initializers.TruncatedNormal(),
                       name='hidden2'))
model.add(layers.Dense(1, activation='sigmoid',
                       name='output'))
model.summary()


layer_outputs = [model.get_layer('hidden1').output,
                 model.get_layer('hidden2').output]
model2 = models.Model(inputs=model.input, outputs=layer_outputs)
model2.summary()


print('####[DNE-05]\nCompile the model using the Adam optimizer, and Cross entroy as a loss function.')
model.compile(optimizer='adam',
              loss='binary_crossentropy',
              metrics=['acc'])


print('####[DNE-06]\nTrain the model.')
time_start = time.perf_counter()
history = model.fit(train_x, train_t,
                    batch_size=len(train_set), epochs=2000, verbose=0)
time_end = time.perf_counter()
print('time elapsed =', time_end-time_start)


print('####[DNE-07]\nPlot charts for the accuracy and loss values.')
DataFrame({'acc': history.history['acc']}).plot(grid=False)
plt.savefig('ex3-4b.png')
if flag_fig: plt.show()
DataFrame({'loss': history.history['loss']}).plot(grid=False)
plt.savefig('ex3-4c.png')
if flag_fig: plt.show()


print('####[DNE-08]\nPlot charts for the final result.')
train_set1 = train_set[train_set['t'] == 1]
train_set2 = train_set[train_set['t'] == 0]

fig = plt.figure(figsize=(7, 7))
subplot = fig.add_subplot(1, 1, 1)
subplot.set_ylim([-15, 15])
subplot.set_xlim([-15, 15])
subplot.scatter(train_set1.x1, train_set1.x2, marker='x')
subplot.scatter(train_set2.x1, train_set2.x2, marker='o')

locations = [[x1, x2]
             for x2 in np.linspace(-15, 15, 100)
             for x1 in np.linspace(-15, 15, 100)]
p_vals = model.predict(np.array(locations)).reshape((100, 100))
subplot.imshow(p_vals, origin='lower', extent=(-15, 15, -15, 15),
               cmap=plt.cm.gray_r, alpha=0.5)
plt.savefig('ex3-4.png')
if flag_fig: plt.show()


h1, h2 = model2.predict(train_x)

z1 = [[[], []], [[], []]]
for h, l in zip(h1, train_t):
    z1[0][l].append(h[0])
    z1[1][l].append(h[1])

plt.title('hidden1')
plt.scatter(z1[0][0], z1[1][0], marker='x')
plt.scatter(z1[0][1], z1[1][1], marker='o')
plt.savefig('ex3-4d.png')
if flag_fig: plt.show()

z2 = [[[], []], [[], []]]
for h, l in zip(h2, train_t):
    z2[0][l].append(h[0])
    z2[1][l].append(h[1])
plt.title('hidden2')
plt.scatter(z2[0][0], z2[1][0], marker='x')
plt.scatter(z2[0][1], z2[1][1], marker='o')
plt.savefig('ex3-4e.png')
if flag_fig: plt.show()
