#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

import numpy as np
import matplotlib.pyplot as plt
import pickle

import tensorflow as tf
from tensorflow.keras import layers, models

from PIL import Image
import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


print('####[OCF-02]\nDownload the ORENIST dataset and store into NumPy arrays.')
with open('ORENIST.data', 'rb') as file:
    images, labels = pickle.load(file)
print('images', type(images), images.shape, type(images[0][0]))
print('labels', type(labels), labels.shape, type(labels[0][0]))

print('####[OCF-03]\nShow some examples.')
fig = plt.figure(figsize=(10, 5))
for i in range(40):
    subplot = fig.add_subplot(4, 10, i+1)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.set_title('%d' % np.argmax(labels[i]))
    subplot.imshow(images[i].reshape(28, 28),
                   vmin=0, vmax=1, cmap=plt.cm.gray_r)
plt.savefig('ex4-1a.png')
if flag_fig: plt.show()


print('####[OCF-04]\nDefine fixed convolutional filters to extract vertical and horizontal edges.')
def edge_filter(shape=None, dtype=None):
    filter0 = np.array(
        [[ 2,  1,  0, -1, -2],
         [ 3,  2,  0, -2, -3],
         [ 4,  3,  0, -3, -4],
         [ 3,  2,  0, -2, -3],
         [ 2,  1,  0, -1, -2]]) / 23.0
    filter1 = np.array(
        [[ 2,  3,  4,  3,  2],
         [ 1,  2,  3,  2,  1],
         [ 0,  0,  0,  0,  0],
         [-1, -2, -3, -2, -1],
         [-2, -3, -4, -3, -2]]) / 23.0

    filter_array = np.zeros([5, 5, 1, 2])
    filter_array[:, :, 0, 0] = filter0
    filter_array[:, :, 0, 1] = filter1

    return filter_array

f = edge_filter()
print('edge_filter:', f.shape)
print(f)
del f


print('####[OCF-05]\nDefine a model to apply the fixed convolutional filters.')
model = models.Sequential(name='TargetModel')
model.add(layers.Reshape((28, 28, 1), input_shape=(28*28,), name='reshape'))
model.add(layers.Conv2D(2, (5, 5), padding='same',
                        kernel_initializer=edge_filter,
                        use_bias=False,
                        name='conv_filter'))
model.add(layers.Lambda(lambda x: abs(x), name='abs'))
model.add(layers.Lambda(lambda x: tf.keras.backend.relu(x-0.2), name='cut_off'))
model.add(layers.MaxPooling2D((2, 2), name='max_pooling'))
print('model')
model.summary()


print('####[OCF-06]\nDefine a model to extract outputs from intermediate layers.')
layer_outputs = [model.get_layer('cut_off').output,
                 model.get_layer('max_pooling').output]
model2 = models.Model(inputs=model.input, outputs=layer_outputs, name='ViewerModel')
print('model2')
model2.summary()

print('####[OCF-07]\nApply the convolutional filters to the ORENST dataset.')
# hm = Image.open('hm.jpg')  # hack an image
# images[0] = np.array(hm.convert('L')).reshape(28*28, )/255.0
conv_output, pool_output = model2.predict(np.array(images[:9]))


print('####[OCF-08]\nShow images after applying the convolutional filters.')
fig = plt.figure(figsize=(10, 3))
filter_vals = edge_filter()
v_max = np.max(conv_output)

for i in range(2):
    subplot = fig.add_subplot(3, 10, 10*(i+1)+1)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(filter_vals[:, :, 0, i], cmap=plt.cm.gray_r)

for i in range(9):
    subplot = fig.add_subplot(3, 10, i+2)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.set_title('%d' % np.argmax(labels[i]))
    subplot.imshow(images[i].reshape((28, 28)),
                   vmin=0, vmax=1, cmap=plt.cm.gray_r)

    subplot = fig.add_subplot(3, 10, 10+i+2)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(conv_output[i, :, :, 0],
                   vmin=0, vmax=v_max, cmap=plt.cm.gray_r)

    subplot = fig.add_subplot(3, 10, 20+i+2)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(conv_output[i, :, :, 1],
                   vmin=0, vmax=v_max, cmap=plt.cm.gray_r)

plt.savefig('ex4-1b.png')
if flag_fig: plt.show()

print('####[OCF-9]\nShow images after applying the pooling layer.')
fig = plt.figure(figsize=(10, 3))
filter_vals = edge_filter()
v_max = np.max(pool_output)

for i in range(2):
    subplot = fig.add_subplot(3, 10, 10*(i+1)+1)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(filter_vals[:, :, 0, i], cmap=plt.cm.gray_r)

for i in range(9):
    subplot = fig.add_subplot(3, 10, i+2)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.set_title('%d' % np.argmax(labels[i]))
    subplot.imshow(images[i].reshape((28, 28)),
                   vmin=0, vmax=1, cmap=plt.cm.gray_r)

    subplot = fig.add_subplot(3, 10, 10+i+2)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(pool_output[i, :, :, 0],
                   vmin=0, vmax=v_max, cmap=plt.cm.gray_r)

    subplot = fig.add_subplot(3, 10, 20+i+2)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(pool_output[i, :, :, 1],
                   vmin=0, vmax=v_max, cmap=plt.cm.gray_r)

plt.savefig('ex4-1c.png')
if flag_fig: plt.show()
