#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

# ORENIST Classification Example

import numpy as np
import matplotlib.pyplot as plt
from pandas import DataFrame
import pickle

import tensorflow as tf
from tensorflow.keras import layers, models

import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


np.random.seed(20190823)
tf.random.set_seed(20190823)


print('####[OCE-02]\nDownload the ORENIST dataset and store into NumPy arrays.')
with open('ORENIST.data', 'rb') as file:
    images, labels = pickle.load(file)


print('####[OCE-03]\nDefine fixed convolutional filters to extract vertical and horizontal edges.')
def edge_filter(shape=None, dtype=None):
    filter0 = np.array(
        [[  2,  1,  0, -1, -2],
         [  3,  2,  0, -2, -3],
         [  4,  3,  0, -3, -4],
         [  3,  2,  0, -2, -3],
         [  2,  1,  0, -1, -2]]) / 23.0
    filter1 = np.array(
        [[  2,  3,  4,  3,  2],
         [  1,  2,  3,  2,  1],
         [  0,  0,  0,  0,  0],
         [ -1, -2, -3, -2, -1],
         [ -2, -3, -4, -3, -2]]) / 23.0

    filter_array = np.zeros([5, 5, 1, 2])
    filter_array[:, :, 0, 0] = filter0
    filter_array[:, :, 0, 1] = filter1

    return filter_array


print('####[OCE-04]\nDefine a model to classify the ORENIST dataset with the fixed filters.')
model = models.Sequential()
model.add(layers.Reshape((28, 28, 1), input_shape=(28*28,), name='reshape'))
model.add(layers.Conv2D(2, (5, 5), padding='same',
                        kernel_initializer=edge_filter,
                        use_bias=False, trainable=False,
                        name='conv_filter'))
model.add(layers.Lambda(lambda x: abs(x), name='abs'))
model.add(layers.Lambda(lambda x: tf.keras.backend.relu(x-0.2), name='cut_off'))
model.add(layers.MaxPooling2D((2, 2), name='max_pooling'))

model.add(layers.Flatten(name='flatten'))
model.add(layers.Dense(2, activation='tanh', name='hidden'))
model.add(layers.Dense(3, activation='softmax', name='softmax'))

model.summary()


print('####[OCE-05]\nCompile the model using the Adam optimizer, and Cross entroy as a loss function.')
model.compile(optimizer='adam',
              loss='categorical_crossentropy',
              metrics=['acc'])


print('####[OCE-06]\nTrain the model.')
time_start = time.perf_counter()
history = model.fit(images, labels,
                    batch_size=len(images), epochs=200, verbose=0)
time_end = time.perf_counter()
print('time elapsed =', time_end-time_start)


print('####[OCE-07]\nPlot a chart to see the accuracy of the model. It achieves the 100% accuracy.')
DataFrame({'acc': history.history['acc']}).plot()
plt.savefig('ex4-2a.png')
if flag_fig: plt.show()


print('####[OCE-08]\nDefine a model to extract outputs from the hidden layer.')
layer_outputs = [model.get_layer('hidden').output]
model2 = models.Model(inputs=model.input, outputs=layer_outputs)


print('####[OCE-09]\nPlot a chart to show the values of $(z_1, z_2)$ for each image.')
hidden_vals = model2.predict(images)

z1_vals = [[], [], []]
z2_vals = [[], [], []]

for hidden_val, label in zip(hidden_vals, labels):
    label_num = np.argmax(label)
    z1_vals[label_num].append(hidden_val[0])
    z2_vals[label_num].append(hidden_val[1])

fig = plt.figure(figsize=(7, 7))
subplot = fig.add_subplot(1, 1, 1)
subplot.scatter(z1_vals[0], z2_vals[0], s=100, marker='|')
subplot.scatter(z1_vals[1], z2_vals[1], s=100, marker='_')
subplot.scatter(z1_vals[2], z2_vals[2], s=100, marker='+')

plt.savefig('ex4-2.png')
if flag_fig: plt.show()
