#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

import numpy as np
from numpy.random import normal
import matplotlib.pyplot as plt

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import layers, models
from tensorflow.python.framework.ops import disable_eager_execution

import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


print('# CNN Interpretation - Image Generation')

disable_eager_execution()

np.random.seed(20190225)
tf.random.set_seed(20190225)


print('####[CII-02]\nDefine a function to generate random images that maximize the outputs from filters.')
def gen_pattern(model_input, layer_ouput, filter_index):
    # ポイント：計算を実行するのか、計算式を登録するのか、区別すること
    # print('gen_pattern()')
    # print('model_input:', type(model_input), model_input.shape)  # Tensor (None, 784)
    # print('layer_output:', type(layer_output), layer_output.shape)  # Tensor (None, 28, 28, 2)
    # print('filter_index:', type(filter_index))  # int
    mean_output = K.mean(layer_output[:, :, :, filter_index])  # 計算式の登録
    # print('mean_output:', type(mean_output), mean_output.shape)  # Tensor ()
    grads = K.gradients(mean_output, model_input)[0]  # 計算式の登録
    # print('grads:', type(grads), grads.shape)  # Tensor (None, 784)
    grads_vals = K.function([model_input], [grads])  # 計算式の登録
    # print('grads_vals:', type(grads_vals), 'inputs:', grads_vals.inputs, 'outputs:',grads_vals.outputs)  # inputs (None, 784) outputs (None, 784)

    input_image = normal(0.5, 0.1, 28*28).reshape([1, 28*28])
    epsilon = 1000
    for _ in range(20):
        input_image += epsilon * grads_vals([input_image])[0]  # 計算の実行
        epsilon *= 0.9

    image = input_image[0]
    image -= np.min(image)
    image /= np.max(image)
    return image


print('####[CII-03]\nDefine a model with fixed filters.')
def edge_filter(shape=None, dtype=None):
    filter0 = np.array(
        [[ 2,  1,  0, -1, -2],
         [ 3,  2,  0, -2, -3],
         [ 4,  3,  0, -3, -4],
         [ 3,  2,  0, -2, -3],
         [ 2,  1,  0, -1, -2]]) / 23.0
    filter1 = np.array(
        [[ 2,  3,  4,  3,  2],
         [ 1,  2,  3,  2,  1],
         [ 0,  0,  0,  0,  0],
         [-1, -2, -3, -2, -1],
         [-2, -3, -4, -3, -2]]) / 23.0

    filter_array = np.zeros([5, 5, 1, 2])
    filter_array[:, :, 0, 0] = filter0
    filter_array[:, :, 0, 1] = filter1

    return filter_array

model = models.Sequential()
model.add(layers.Reshape((28, 28, 1), input_shape=(28*28,), name='reshape'))
model.add(layers.Conv2D(2, (5, 5), padding='same',
                        kernel_initializer=edge_filter,
                        use_bias=False,
                        name='conv_filter'))
model.add(layers.Lambda(lambda x: abs(x), name='abs'))
model.add(layers.Lambda(lambda x: tf.keras.backend.relu(x-0.2), name='cut_off'))
model.summary()


print('####[CII-04]\nGenerate images that maximize the outputs from the fixed filters.')
fig = plt.figure(figsize=(2, 1))
layer_output = model.get_layer('cut_off').output
for i in range(layer_output.shape[-1]):
    img = gen_pattern(model.input, layer_output, i)
    subplot = fig.add_subplot(1, 2, i+1)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(img.reshape(28, 28), cmap=plt.cm.gray_r)
plt.savefig('ex5-3a.png')
if flag_fig: plt.show()

exit()


print('####[CII-07]\nRestore the model from the exported model file.')
cnn_model = models.load_model('MNIST_double.hd5')
cnn_model.summary()


print('####[CII-08]\nGenerate images that maximize the outputs from the first filters.')
fig = plt.figure(figsize=(16, 2))
layer_output = cnn_model.get_layer('conv_filter1').output
for i in range(layer_output.shape[-1]):
    img = gen_pattern(cnn_model.input, layer_output, i)
    subplot = fig.add_subplot(2, 16, i+1)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(img.reshape(28, 28), cmap=plt.cm.gray_r)
plt.savefig('ex5-3b.png')
if flag_fig: plt.show()


print('####[CII-09]\nGenerate images that maximize the outputs from the second filters.')
fig = plt.figure(figsize=(16, 4))
layer_output = cnn_model.get_layer('conv_filter2').output
for i in range(layer_output.shape[-1]):
    img = gen_pattern(cnn_model.input, layer_output, i)
    subplot = fig.add_subplot(4, 16, i+1)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(img.reshape(28, 28), cmap=plt.cm.gray_r)
plt.savefig('ex5-3c.png')
if flag_fig: plt.show()
