#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt

import tensorflow as tf
from tensorflow.keras import layers, models, initializers
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.datasets import cifar10


import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


print('# CIFAR10 CNN Classification')

np.random.seed(20190610)
tf.random.set_seed(20190610)

print('####[CCC-02]\nDownload the CIFAR10 dataset and store into NumPy arrays.')

# Windows でこの↓エラーが出るので
# Exception: URL fetch failure on https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz: None -- [SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: certificate has expired (_ssl.c:1129)
# この↓ファイルをハックした
# c:/Users/t-izu/AppData/Local/Programs/Python/Python39/Lib/site-packages/keras/datasets/cifar10.py
(train_images, train_labels), (test_images, test_labels) = cifar10.load_data()
print('original data')
print('train_images:', type(train_images), train_images.shape)
print('train_labels:', type(train_labels), train_labels.shape)
print('test_images:',  type(test_images),  test_images.shape)
print('test_labels:',  type(test_labels),  test_labels.shape)

train_images = train_images.astype('float32') / 255
test_images = test_images.astype('float32') / 255
train_labels = tf.keras.utils.to_categorical(train_labels, 10)
test_labels = tf.keras.utils.to_categorical(test_labels, 10)

print('reshaped data')
print('train_images:', type(train_images), train_images.shape)
print('train_labels:', type(train_labels), train_labels.shape)
print('test_images:',  type(test_images),  test_images.shape)
print('test_labels:',  type(test_labels),  test_labels.shape)


print('####[CCC-03]\nShow sample images for each category.')
fig = plt.figure(figsize=(8, 10))
c = 0
for i in range(10):
    for j in range(len(train_images)):
        if np.argmax(train_labels[j]) == i:
            c += 1
            subplot = fig.add_subplot(10, 8, c)
            subplot.set_xticks([])
            subplot.set_yticks([])
            subplot.imshow(train_images[j])
            if c % 8 == 0:
                break
plt.savefig('ex5-5a.png')
if flag_fig: plt.show()


print('####[CCC-04]\nDefine the image data generator.')
datagen = ImageDataGenerator(
    rotation_range=10,
    width_shift_range=0.1,
    height_shift_range=0.1,
    zoom_range=[0.8, 1.2],
    horizontal_flip=True,
    channel_shift_range=0.2)

if False:
    datagen_flow = datagen.flow(np.array(train_images), batch_size=7)
    print('datagen_flow:', type(datagen_flow))
    datagen_flow_0 = datagen_flow[0]
    print('datagen_flow_0:', type(datagen_flow_0), datagen_flow_0.shape)

    fig = plt.figure(figsize=(9, 7))
    for i in range(9):
        datagen_flow_0 = next(datagen_flow)
        print('datagen_flow_0:', type(datagen_flow_0), datagen_flow_0.shape)
        for j in range(7):
            subplot = fig.add_subplot(7, 9, i+9*j+1)
            subplot.set_xticks([])
            subplot.set_yticks([])
            subplot.imshow(datagen_flow_0[j])

    if flag_fig: plt.show()
    del datagen_flow, datagen_flow_0


print('####[CCC-05]\nShow some generated images for each category.')
fig = plt.figure(figsize=(8, 10))
c = 0
for i in range(10):
    for j in range(len(train_images)):
        if np.argmax(train_labels[j]) == i:
            break
    c += 1
    subplot = fig.add_subplot(10, 8, c)
    subplot.set_xticks([])
    subplot.set_yticks([])
    subplot.imshow(train_images[j])

    for _ in range(7):
        img = datagen.flow(np.array([train_images[j]]), batch_size=1)[0][0]
        c += 1
        subplot = fig.add_subplot(10, 8, c)
        subplot.set_xticks([])
        subplot.set_yticks([])
        subplot.imshow(img)
plt.savefig('ex5-5b.png')
if flag_fig: plt.show()


print('####[CCC-06]\nDefine a CNN model.')
model = models.Sequential()
model.add(layers.Conv2D(32, (3, 3), padding='same',
                        kernel_initializer=initializers.TruncatedNormal(),
                        use_bias=True, activation='relu',
                        input_shape=(32, 32, 3),
                        name='conv_filter1-1'))
model.add(layers.Conv2D(32, (3, 3), padding='same',
                        kernel_initializer=initializers.TruncatedNormal(),
                        use_bias=True, activation='relu',
                        name='conv_filter1-2'))
model.add(layers.MaxPooling2D((2, 2), name='max_pooling1'))
model.add(layers.Dropout(rate=0.25, name='dropout1'))

model.add(layers.Conv2D(64, (3, 3), padding='same',
                        kernel_initializer=initializers.TruncatedNormal(),
                        use_bias=True, activation='relu',
                        name='conv_filter2-1'))
model.add(layers.Conv2D(64, (3, 3), padding='same',
                        kernel_initializer=initializers.TruncatedNormal(),
                        use_bias=True, activation='relu',
                        name='conv_filter2-2'))
model.add(layers.MaxPooling2D((2, 2), name='max_pooling2'))
model.add(layers.Dropout(rate=0.25, name='dropout2'))

model.add(layers.Flatten(name='flatten'))
model.add(layers.Dense(512, activation='relu',
                       kernel_initializer=initializers.TruncatedNormal(),
                       name='hidden'))
model.add(layers.Dropout(rate=0.5, name='dropout3'))
model.add(layers.Dense(10, activation='softmax', name='softmax'))

model.summary()


print('####[CCC-07]\nCompile the model using the Adam optimizer, and the cross entroy as a loss function.')
model.compile(optimizer='adam',
              loss='categorical_crossentropy',
              metrics=['acc'])


print('####[CCC-08]\nTrain the model. It achieves the 82% accuracy.')
batch_size = 64
time_start = time.perf_counter()
history = model.fit_generator(
    datagen.flow(train_images, train_labels, batch_size=batch_size),
    validation_data=(test_images, test_labels),
    steps_per_epoch=len(train_images) / batch_size,
    epochs=20)
time_end = time.perf_counter()
print('time elapsed =', time_end-time_start)


print('####[CCC-09]\nPlot chars for accuracy and loss values.')
DataFrame({'acc': history.history['acc'],
           'val_acc': history.history['val_acc']}).plot()
plt.savefig('ex5-5c.png')
if flag_fig: plt.show()
DataFrame({'loss': history.history['loss'],
           'val_loss': history.history['val_loss']}).plot()
plt.savefig('ex5-5d.png')
if flag_fig: plt.show()

print('####\nExport the trained model as a file `CIFAR10.hd5`')
model.save('CIFAR10.hd5', save_format='h5')
