#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import layers, models
from tensorflow.keras.datasets import mnist

import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


print('# MNIST Autoencoder Example')


print('####[MAE-02]\nDownload the MNIST dataset and store into NumPy arrays.')
(train_images, train_labels), (test_images, test_labels) = mnist.load_data()
train_images = train_images.reshape(
    (len(train_images), 784)).astype('float32') / 255
test_images = test_images.reshape(
    (len(test_images), 784)).astype('float32') / 255
train_labels = tf.keras.utils.to_categorical(train_labels, 10)
test_labels = tf.keras.utils.to_categorical(test_labels, 10)


print('####[MAE-03]\nDefine an autoencoder model for MNIST images.')
model = models.Sequential()
model.add(layers.Dense(256, activation='relu', input_shape=(28*28,), name='enc256'))
model.add(layers.Dense(128, activation='relu', name='enc128'))
model.add(layers.Dense(16, activation='relu', name='enc16'))
model.add(layers.Dense(128, activation='relu', name='dec128'))
model.add(layers.Dense(256, activation='relu', name='dec256'))
model.add(layers.Dense(784, activation='sigmoid', name='dec784'))

model.summary()


print('####[MAE-04]\nCompile the model using the Adam optimizer, and MSE as a loss function.')
model.compile(optimizer='adam', loss='mse')


print('####[MAE-05]\nDefine a function to show reconstructed images.')
def show_result():
    results = model.predict(test_images[:8])

    fig = plt.figure(figsize=(16, 2.7))
    for i in range(8):
        subplot = fig.add_subplot(2, 10, i+1)
        subplot.set_xticks([])
        subplot.set_yticks([])
        subplot.imshow(test_images[i].reshape((28, 28)),
                       vmin=0, vmax=1, cmap=plt.cm.gray_r)

        subplot = fig.add_subplot(2, 10, i+11)
        subplot.set_xticks([])
        subplot.set_yticks([])
        subplot.imshow(results[i].reshape((28, 28)),
                       vmin=0, vmax=1, cmap=plt.cm.gray_r)


print('####[MAE-06]\nTrain the model for a single epoch with 30,000 images, and show the intermediate result.')
history = model.fit(train_images[:30000], train_images[:30000],
                    batch_size=256, epochs=1)
show_result()
plt.savefig('ex5-6a.png')
if flag_fig: plt.show()


print('####[MAE-07]\nTrain the model for additional epochs, and show the final result.')
time_start = time.perf_counter()
history = model.fit(train_images, train_images,
                    batch_size=256, epochs=6)
time_end = time.perf_counter()
print('time elapsed =', time_end-time_start)
show_result()
plt.savefig('ex5-6b.png')
if flag_fig: plt.show()

DataFrame({'loss': history.history['loss']}).plot()
plt.savefig('ex5-6c.png')
if flag_fig: plt.show()

print('####\nExport the trained model as a file `MNIST_autoenc.hd5`')
model.save('MNIST_autoenc.hd5', save_format='h5')


print('####\nShow encoded vectors')
get_enc16_out = K.function([model.input], [model.get_layer('enc16').output])
get_dec784_out = K.function([model.input], [model.get_layer('dec784').output])
for i in range(10):
    cmax = 8
    fig = plt.figure(figsize=(6, cmax))
    c = 0
    for j in range(len(test_images)):
        if np.argmax(test_labels[j]) == i:
            img0 = np.array(test_images[j])
            subplot = fig.add_subplot(cmax, 3, c*3+1)
            subplot.set_xticks([])
            subplot.set_yticks([])
            subplot.imshow(img0.reshape(28, 28), cmap=plt.cm.gray_r)
            (img1, ) = get_dec784_out([img0.reshape(1, 28*28)])
            subplot = fig.add_subplot(cmax, 3, c*3+2)
            subplot.set_xticks([])
            subplot.set_yticks([])
            subplot.imshow(img1[0].reshape(28, 28), cmap=plt.cm.gray_r)
            (enc16, ) = get_enc16_out([img0.reshape(1, 28*28)])
            print('enc16:', type(enc16), len(enc16), type(enc16[0]), enc16[0].shape, enc16[0])
            subplot = fig.add_subplot(cmax, 3, c*3+3)
            subplot.set_xticks([])
            subplot.set_xlim(-0.5, 15.5)
            subplot.bar(range(16), enc16[0], align='center')
            c += 1
        if c >= cmax: break
    plt.savefig('ex5-6x{}.png'.format(str(i)))
    if flag_fig: plt.show()
