#!/usr/bin/python3
# -*- coding: utf-8 -*-

# The code is derived from https://github.com/enakai00/colab_tfbook
# and modified by izumi@ieee.org, 2021

# このコードはこの↓書籍のサンプルコードをベースに改変したものである
# 『TensorFlow と Keras で 動かしながら学ぶ ディープラーニングの仕組み』
# 中井悦司, マイナビ出版, 2019年
# https://book.mynavi.jp/ec/products/detail/id=109925

import numpy as np
import matplotlib.pyplot as plt

import tensorflow as tf
from tensorflow.keras import layers, models, initializers
from tensorflow.keras.datasets import mnist

import time
import sys
import os
import specrep
print(*sys.argv)
if 'nogpu' in sys.argv:
    # tf.device('/CPU:0')
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
else:
    # tf.device('/GPU:0')
    pass
flag_fig  = False if 'nofig'  in sys.argv else True
flag_spec = False if 'nospec' in sys.argv else True
if flag_spec: specrep.report()
if (not tf.__version__.startswith('2.')):
    print('This code requires version 2.*.')
    exit()


print('# MNIST DCGAN Example')

np.random.seed(20191019)
tf.random.set_seed(20191019)


print('####[MDE-02]\nDownload the MNIST dataset and store into NumPy arrays.')
(train_images, train_labels), (test_images, test_labels) = mnist.load_data()
train_images = train_images.reshape(
    (len(train_images), 784)).astype('float32') / 255
test_images = test_images.reshape(
    (len(test_images), 784)).astype('float32') / 255
train_labels = tf.keras.utils.to_categorical(train_labels, 10)
test_labels = tf.keras.utils.to_categorical(test_labels, 10)


print('####[MDE-03]\nDefina a generator model.')
latent_dim = 64
generator = models.Sequential()
generator.add(
    layers.Dense(7*7*128, kernel_initializer=initializers.TruncatedNormal(),
                 input_shape=(latent_dim,), name='expand'))
generator.add(layers.LeakyReLU(name='leaky_relu1'))
generator.add(layers.Reshape((7, 7, 128), name='reshape'))
generator.add(
    layers.Conv2DTranspose(64, 5, strides=2, padding='same',
                           kernel_initializer=initializers.TruncatedNormal(),
                           name='deconv1'))
generator.add(layers.LeakyReLU(name='leaky_relu2'))
generator.add(
    layers.Conv2DTranspose(1, 5, strides=2, padding='same',
                           kernel_initializer=initializers.TruncatedNormal(),
                           activation='sigmoid', name='deconv2'))
generator.add(layers.Flatten(name='flatten'))
generator.summary()


print('####[MDE-04]\nDefina a discriminator model.')
discriminator = models.Sequential()
discriminator.add(layers.Reshape((28, 28, 1), input_shape=((28*28,)),
                                 name='reshape'))
discriminator.add(
    layers.Conv2D(64, (5, 5), strides=2, padding='same',
                  kernel_initializer=initializers.TruncatedNormal(),
                  name='conv1'))
discriminator.add(layers.LeakyReLU(name='leaky_relu1'))
discriminator.add(
    layers.Conv2D(128, (5, 5), strides=2, padding='same',
                  kernel_initializer=initializers.TruncatedNormal(),
                  name='conv2'))
discriminator.add(layers.LeakyReLU(name='leaky_relu2'))
discriminator.add(layers.Flatten(name='flatten'))
discriminator.add(layers.Dropout(rate=0.4, name='dropout'))
discriminator.add(layers.Dense(1, activation='sigmoid', name='sigmoid'))
discriminator.summary()


print('####[MDE-05]\nCompile the discriminator using the Adam optimizer, and Cross entroy as a loss function.')
discriminator.compile(optimizer='adam', loss='binary_crossentropy')


print('####[MDE-06]\nDefine an end-to-end GAN model to train the generator.')
discriminator.trainable = False
# fit や train_on_batch などで学習を行うとき、
# trainable = False の層は逆伝播の重みの更新は停止されるものの、
# 順伝播時の平均・分散の更新は続くために出力値が変化する。
# trainable属性の変更後は compile() が必要。
gan_input = tf.keras.Input(shape=(latent_dim,))
gan_output = discriminator(generator(gan_input))
gan_model = models.Model(gan_input, gan_output)
print('gan_input:', type(gan_input))
print('gan_output:', type(gan_output))
print('gan_model:', type(gan_model))
gan_model.summary()


print('####[MDE-07]\nCompile the GAN model using the Adam optimizer, and Cross entroy as a loss function.')
gan_model.compile(optimizer='adam', loss='binary_crossentropy')


print('####[MDE-08]\nDefine some working variables to trace the training process.')
batch_size = 32
image_num = 0
step = 0
examples = []
sample_inputs = np.random.rand(8, latent_dim) * 2.0 - 1.0
examples.append(generator.predict(sample_inputs))


print('####[MDE-09]\nTrain the model for 40,000 batches.')
time_start = time.perf_counter()
for _ in range(40000):
    random_inputs = np.random.rand(batch_size, latent_dim) * 2.0 - 1.0
    generated_images = generator.predict(random_inputs)
    real_images = train_images[image_num:image_num+batch_size]
    all_images = np.concatenate([generated_images, real_images])
    labels = np.concatenate([np.zeros((batch_size, 1)),
                             np.ones((batch_size, 1))])
    labels += 0.05 * np.random.random(labels.shape)  # ???
    # print('all_images:', type(all_images), all_images.shape)
    # print('labels:', type(labels), labels.shape)
    # discriminator.trainable==False だが、変更前にcompileしていることに注意
    d_loss = discriminator.train_on_batch(all_images, labels)

    random_inputs = np.random.rand(batch_size, latent_dim) * 2.0 - 1.0
    fake_labels = np.ones((batch_size, 1))
    g_loss = gan_model.train_on_batch(random_inputs, fake_labels)

    image_num += batch_size
    if image_num + batch_size > len(train_images):
        image_num = 0

    step += 1
    if step % 400 == 0:
        print('step: {:7}, loss(discriminator, generator): {:6.4f}, {:6.4f}'.format(step, d_loss, g_loss))
    if step % 4000 == 0:
        examples.append(generator.predict(sample_inputs))
time_end = time.perf_counter()
print('time elapsed =', time_end-time_start)


print('####[MDE-10]\nShow the progress of sample images.')
def show_images(examples):
    fig = plt.figure(figsize=(10, 1.2*len(examples)))
    c = 1
    for images in examples:
        for image in images:
            subplot = fig.add_subplot(len(examples), 8, c)
            subplot.set_xticks([])
            subplot.set_yticks([])
            subplot.imshow(image.reshape((28, 28)),
                           vmin=0, vmax=1, cmap=plt.cm.gray_r)
            c += 1

show_images(examples)
plt.savefig('ex5-7.png')
if flag_fig: plt.show()


print('####[MDE-11]\nExport the trained model as a file `gan_generator.hd5`.')
generator.save('gan_generator.hd5', save_format='h5')
